/* 
* Rootfly:  A program for analyzing and measuring roots. 
* Copyright (c) 2007 Clemson University.
*
* This program is free software; you can redistribute it and/or modify
* it under the terms of the GNU General Public License as published by
* the Free Software Foundation; either version 2 of the License, or (at
* your option) any later version.
* 
* This program is distributed in the hope that it will be useful, but
* WITHOUT ANY WARRANTY; without even the implied warranty of
* MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU
* General Public License for more details.
* 
* You should have received a copy of the GNU General Public License
* along with this program; if not, write to the Free Software
* Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
*/ 

////////////////////////////////////////////////////////
//  This file contains the following classes:
//    Subject (Note:  RootflyDlg inherits from Observer)
//    Observer
//    Controller, which inherits from Subject



#ifndef CONTROLLER_H
#define CONTROLLER_H

#include <list>
#include <iostream>
#include <vector>
#include "DataStructures.h"
#include "Command.h"
#include "stdafx.h"

 

/// Notification codes that will be sent through Notify -> OnUpdate
enum notification_codes 
{ 
ON_INIT, 
ROOT_ADDED, 
ROOT_DELETED, 
ROOT_RESIZED,
ROOT_INSTANCE_ADDED, 
POINT_ADDED, 
POINT_REMOVED, 
POINT_CHANGED, 
POINT_MOVED,
DIAMETER_ADDED, 
DIAMETER_REMOVED, 
DIAMETER_CHANGED,
DIAMETER_MOVED, 
SESSION_CHANGED,
TUBE_CHANGED,
EXPERIMENT_CHANGED,
CURRENT_IMAGE_CHANGED,
SESSION_DATE_CHANGED,
OPERATOR_CHANGED,
SAVE_STATUS_SAVED,
SAVE_STATUS_UNSAVED,
FILENAME_CHANGED, 
DEFAULT_DIR_CHANGED,
IMAGEDIR_CHANGED,
IMAGEFILEFORMAT_CHANGED,
WINDOW_CHANGED, 
COLOR_CHANGED, 
STATUS_CHANGED, 
CONDITION_CODE_ADDED, 
CONDITION_CODE_SELECTED,
CONDITION_SET_CHANGED,
CONDITION_DEFAULT_INDEX_CHANGED,
SHOW_ROOT_ID_CHANGED, 
SHOW_DEAD_ROOTS_CHANGED, 
SHOW_DIAMETERS_CHANGED, 
DRAW_SETTINGS_CHANGED, 
TUBE_ADDED,
NEW_FILE_LOADED,
FILE_SAVED,
ROOT_INSTANCE_CLEAR,
ROOT_CLEAR,
TUBE_CLEAR,
IMAGED_SCANNED,
IMAGE_ANNOTATED,
EXPERIMENT_CLEAR,
PREFERENCES_READ,
PREFERENCES_SAVED,
MIN_WINDOWS_CHANGED,
MIN_SESSIONS_CHANGED,
MAX_WINDOWS_CHANGED,
MAX_SESSIONS_CHANGED,
RD_THUMBNAILS,
RD_OVERLAY,
RD_ALL,
SESSION_NUMBER_CHANGED,
MM_CONVERSION_CHANGED,
ROOT_DRAWING_COLOR_CHANGED,
SELECTED_ROOT_DRAWING_COLOR_CHANGED,
DEAD_SELECTED_ROOT_DRAWING_COLOR_CHANGED,
COPY_PREVIOUS,
SESSION_DATES_SET,
UNDO,
REDO
};

using namespace std ;

class Subject;

/**@class Observer
@brief Attached to the subject so that they can be notified on changes
       to the data structures though OnUpdate
*/
class Observer {
public:
  Observer() {};
  ~Observer() {};
  /// Virtual function defined in RootflyDlg.h - Called on Notify
  virtual void OnUpdate(Subject* theChangeSubject, int code_update) = 0;
};


/**@class Subject
@brief Updates all observers when changes in the data structures are made via
       the notify function which will create an event through OnUpdate
*/
class Subject {
public:
  Subject() {};
  ~Subject() {};
  void Attach(Observer*); ///> Attach an observer to the subject
  void Detach(Observer*); ///> Detach an observer from the subject
  void Notify(int code_notify); ///> Notify observers of a change
private:
  vector<Observer*> m_observers; ///> Observers of the data structures
};

/**@class Controller
@brief The purpose of the Controller class is to provide a clean separation between
  the data and the display.  The display get read-only access to the data
  through the controller, but in order to write to the data it has to call
  one of the controller wrapper functions.

  The Controller also contains functionality to send notification messages
  to the RootflyDlg message handler whenever data is changed.  This functionality
  is currently only being used in a limited way, to update the title bar.
*/
class Controller : public Subject {
public:
  Controller()
  { 
  }
  ~Controller() 
  {
  }
  
  //Functions that return pointers may also return NULL.

  const RootflyData* GetRootflyData() const { return &m_data; } ///> Returns const pointer to RootflyData

  void C_CreateNewProject(const CString& filename, const CString& directory); ///> Wrapper for creating a new project

  void C_SetTube(const CString& tube); ///> Wrapper for setting tube number
  void C_SetWindow(int window); ///> Wrapper for setting window number
  void C_SetSession(int session); ///> Wrapper for setting session number
  
  void C_SetExperiment(const CString& experiment); ///> Wrapper for setting experiment
  void C_SetCurrentImage(const CString& current_image); ///> Wrapper for setting current image
  void C_SetSessionDate(const CString& date); ///> Wrapper for setting session data
  void C_SetOperator(const CString& operator_initials); ///> Wrapper for setting operator
  void C_SetFilename(const CString& filename); ///> Wrapper for setting filename

  void C_SetImageDir(const CString& imgdir); ///> Wrapper for setting image directory
  void C_SetImageFileFormat(ImageFileFormat iff); ///> Wrapper for setting image file format
  ImageFileFormat C_GetImageFileFormat() const; ///> Wrapper for getting image file format

  // versions that use the current session
  RPoint* C_CreatePoint(int root_id); ///> Wrapper for creating a point
  RPoint* C_CreateMidPoint(int root_id, int index); //> Wrapper for creating a mid point
  bool C_DeletePoint(int index, int root_id); ///> Wrapper for deleting a point
  bool C_ExtendEndpoint(RootInstance &old_ri, int root_id, int session); // used for undo/redo extend endpoint
  bool C_MovePoint(int index, int root_id, int x, int y);  ///> Wrapper for moving a point
  bool C_MovePoint_U(int index, int root_id, int x, int y, RootInstance * prev);  ///> Wrapper for moving a point
  bool C_ChangeWindowOfPoint(int index, int root_id, int w); ///> Wrapper for changing the window of a point
  bool C_MoveRootsInWindow(int window, int offset_x, int offset_y); ///> Wrapper for moving all roots in a window in current session
  bool C_MoveRootsInWindow_U(int window, int offset_x, int offset_y, int total_offset_x, int total_offset_y);
  RDiameter* C_CreateDiameter(int root_id); ///> Wrapper for creating a diameter
  bool C_DeleteDiameter(int index, int root_id); ///> Wrapper for deleting a diameter
  bool C_MoveDiameter(int index, int root_id, const RPoint& r);  ///> Wrapper for moving a diameter (current session)
  bool C_MoveDiameter_U(int index, int root_id, const RPoint& old_r, const RPoint& new_r);
  bool C_MoveDiameter(int index, int root_id, int x, int y);  ///> Wrapper for moving a diameter (current session)
  bool C_ChangeWindowOfDiameter(int index, int root_id, int w); ///> Wrapper for changing the window of a diameter
  bool C_ResizeDiameter(int index, int root_id, double radius);  ///> Wrapper for resizing a diameter
  bool C_ResizeDiameter_U(int index, int root_id, double old_radius, double new_radius);
  RootInstance* C_CreateRootInstance(int root_id);  ///> Wrapper for creating a root instance
//  void C_SetSessionNumber(int session_number, int root_id); ///> Wrapper for setting session number
  Root* C_CreateRoot(int primaryWindow); ///> Wrapper for creating a root
  void C_CreateRoot(int primaryWindow, const RootInstance& ri, CString default_cond); ///> Wrapper for creating a root
  Tube* C_CreateTube(const CString& tag); ///> Wrapper for creating a tube
  bool C_DeleteRoot(int index); ///> Wrapper for deleting a root
  
  // versions with session specified
  bool C_MoveRootSess(int root_id, int session, int offset_x, int offset_y); ///> Wrapper for moving a root
  bool Controller::C_MoveRootSessU(int root_id, int session, int offset_x, int offset_y, int total_offset_x, int total_offset_y);
  void C_MovePointSess(int index, int root_id, int session, int x, int y);  ///> Wrapper for moving a point
  bool C_MoveDiameterSess(int index, int root_id, int session, int x, int y);  ///> Wrapper for moving a diameter
  bool C_ResizeDiameterSess(int index, int root_id, int session, double radius);  ///> Wrapper for resizing a diameter

  void C_CopySession(int from, int to, const CString &tubename, int curr_window);  ///> Copies all roots existing in session "from" to a given session.
  
  void C_SetShowRootID(bool status); ///> Wrapper for setting show root ids
  void C_SetShowDeadRoots(bool status); ///> Wrapper for setting show dead roots
  void C_SetShowDiameters(bool status); ///> Wrapper for setting show diameters
  
  int C_LoadFile(const CString& fname); ///> Wrapper for loading a file
  int C_SaveFile(const CString& fname); ///> Wrapper for saving a file
  
  bool C_ClearAllRootInstance(int root_id); ///> Calls ClearAll for the root instance in the current TWS
  bool C_ClearAllRoot(int root_id); ///> Calls ClearAll for the root in the current TWS
  bool C_ClearAllTube(); ///> Calls ClearAll for the tube
  bool C_ClearAllExperiment(); ///> Calls ClearAll for the experiment
  
  void C_ReadUserPreferences(); ///> Wrapper to read settings from registry
  void C_SaveUserPreferences(); ///> Wrapper to save settings to registry

  void C_SetMinWindows(int windows); ///> Wrapper to set min number of windows
  void C_SetMinSessions(int sessions); ///> Wrapper to set min number of sessions
  void C_SetMaxWindows(int windows); ///> Wrapper to set max number of windows
  void C_SetMaxSessions(int sessions); ///> Wrapper to set max number of sessions
  void C_SetMMConversion(float conversion); ///> Wrapper to set mm conversion

  void C_SetAutoDetectOptions(BOOL on_load, BOOL check_diameter, float diameter_value, BOOL check_length, 
                                         float length_value, BOOL check_vertical, float degree_value);


  void C_SetRootDrawingColor(COLORREF color); ///> Wrapper to set root drawing color
  void C_SetRootDrawingColorSelected(COLORREF color); ///> Wrapper to set selected root drawing color
  void C_SetRootDrawingColorDead(COLORREF color); ///> Wrapper to set dead root drawing color
  void C_SetRootDrawingColorDeadSelected(COLORREF color); ///> Wrapper to set dead selected root drawing color

  void C_SetRootColor(int root_id, int color_choice); ///> Wrapper to set color choice for the root
  void C_SetRootAlive(int root_id, bool alive); ///> Wrapper to set the root alive or dead
  void C_SetConditionCode(int root_id, CString condition); ///> Wrapper to set the root condition code
  void C_SetConditionCode(int root_id, CString condition, bool undo); ///> Wrapper to set the root condition code (and make undoable)
  void C_AddConditionCode(CString condition); ///> Wrapper to add a condition code to list
  void C_SetConditionSet(std::vector<CString> set); ///> Wrapper to to set the condition code vector
  void C_SetConditionDefaultIndex(int index); ///> Wrapper to to set the default condition code index

  void C_SetFileDirectory(CString dir); ///> Wrapper to to set default file directory
  void C_Export(CString filename);///Helper function for rootflydlg export

  void C_ConvertOldRoots(const CString& dir, ImageFileFormat& image_file_format);
  void C_ConvertOldRoots(const CString& dir, ImageFileFormat& image_file_format, CWnd *pParentWnd); // has progress bar

  void C_SetSessionDates(std::vector<Date> dates); ///> Wrapper to set the session dates
  void C_SetImageScanned(bool status); ///> Wrapper to set the current image to a false positive
  void C_SetImageAnnotated(bool status); 
  void C_SetModify(bool modify); ///> Wrapper for setting save status

  // undo/redo funcitons
  bool DoCommand(Command* com);
  void Undo();
  void Redo();
  void ClearRedoList();
  void ClearUndoList();
  void ClearUndoRedo();
  bool isUndoVectorEmpty();
  bool isRedoVectorEmpty();

private:
  int CalculateDaysBetweenDates(Date start, Date end);
  RootflyData m_data; ///> Main data structure
  RootInstance *C_GetRootInstance(int root_id); ///> Gets a root instance from a root ID number
  RootInstance *C_GetRootInstance(int root_id, int session); ///> Gets a root instance from a root ID number and session number

  //undo/redo variables

  typedef std::vector<Command*> CommandList;
  CommandList m_undolist;
  CommandList m_redolist; 
  void ClearCommandList(CommandList* pList);

};


#endif
